# ------------------------------------------------------------------------------
# Clorm Query implementation
# ------------------------------------------------------------------------------

from __future__ import annotations

import functools
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from clorm.orm.factcontainers import FactMap

from ._typing import _T0, _T1, _T2, _T3, _T4
from .core import Predicate, and_
from .query import (
    Query,
    QueryExecutor,
    QuerySpec,
    make_query_plan,
    process_join,
    process_orderby,
    process_where,
)

# ------------------------------------------------------------------------------
# Global
# ------------------------------------------------------------------------------
SelfQuery = TypeVar("SelfQuery", bound="BaseQueryImpl[Any]")
_T = TypeVar("_T", bound=Any)
_Fn = TypeVar("_Fn", bound=Callable[..., Any])

# ------------------------------------------------------------------------------
# New Clorm Query API
#
# QueryImpl
# - factmaps             - dictionary mapping predicate types to FactMap objects
# - qspec                - a dictionary with query parameters
# ------------------------------------------------------------------------------


def _generate(fn: _Fn) -> _Fn:
    """Copy decorator.

    This decorator copies the object and runs the copied object.
    """

    @functools.wraps(fn)
    def wrap(self: "BaseQueryImpl", *args: Any, **kw: Any) -> Any:
        self = self.__class__(self._factmaps, self._qspec)
        x = fn(self, *args, **kw)
        assert x is self, "methods must return self"
        return self

    return wrap  # type: ignore


@overload
def _check_join_called_first(*, endpoint: bool = False) -> Callable[[_Fn], _Fn]: ...


@overload
def _check_join_called_first(_fn: _Fn, *, endpoint: bool = False) -> _Fn: ...


def _check_join_called_first(_fn=None, *, endpoint=False):
    """test whether a precondition to call the decorated function is met"""

    def wrap(fn: _Fn) -> _Fn:
        @functools.wraps(fn)
        def check(self: "BaseQueryImpl", *args: Any, **kwargs: Any) -> Any:
            if self._qspec.join is not None or len(self._qspec.roots) == 1:
                return fn(self, *args, **kwargs)
            if endpoint:
                raise ValueError(
                    (
                        "A query over multiple predicates is incomplete without "
                        "'join' clauses connecting these predicates"
                    )
                )
            raise ValueError("A 'join' clause must be specified before '{}'".format(fn.__name__))

        return check  # type: ignore

    if _fn is None:
        return wrap
    return wrap(_fn)


class BaseQueryImpl(Query, Generic[_T]):
    def __init__(self, factmaps: Dict[Type[Predicate], FactMap], qspec: QuerySpec) -> None:
        self._factmaps = factmaps
        self._qspec = qspec

    # --------------------------------------------------------------------------
    # Add a join expression
    # --------------------------------------------------------------------------
    @_generate
    def join(self: SelfQuery, *expressions: Any) -> SelfQuery:
        join = process_join(expressions, self._qspec.roots)
        self._qspec = self._qspec.newp(join=join)
        return self

    # --------------------------------------------------------------------------
    # Add an order_by expression
    # --------------------------------------------------------------------------
    @_generate
    @_check_join_called_first
    def where(self: SelfQuery, *expressions: Any) -> SelfQuery:
        if not expressions:
            self._qspec.newp(where=None)  # Raise an error

        if len(expressions) == 1:
            where = process_where(expressions[0], self._qspec.roots)
        else:
            where = process_where(and_(*expressions), self._qspec.roots)

        self._qspec = self._qspec.newp(where=where)
        return self

    # --------------------------------------------------------------------------
    # Add an orderered() flag
    # --------------------------------------------------------------------------
    @_generate
    @_check_join_called_first
    def ordered(self: SelfQuery, *expressions: Any) -> SelfQuery:
        if self._qspec.getp("order_by", None) is not None:
            raise ValueError(
                (
                    "Invalid query 'ordered' declaration conflicts "
                    "with previous 'order_by' declaration"
                )
            )
        self._qspec = self._qspec.newp(ordered=True)
        return self

    # --------------------------------------------------------------------------
    # Add an order_by expression
    # --------------------------------------------------------------------------
    @_generate
    @_check_join_called_first
    def order_by(self: SelfQuery, *expressions: Any) -> SelfQuery:
        if not expressions:
            self._qspec = self._qspec.newp(order_by=None)  # raise exception
        elif self._qspec.getp("ordered", False):
            raise ValueError(
                (
                    "Invalid query 'order_by' declaration '{}' "
                    "conflicts with previous 'ordered' "
                    "declaration"
                ).format(expressions)
            )
        else:
            self._qspec = self._qspec.newp(
                order_by=process_orderby(expressions, self._qspec.roots)
            )
        return self

    @_generate
    @_check_join_called_first
    def select(self, *outsig: Any) -> Any:
        if not outsig:
            raise ValueError("An empty 'select' signature is invalid")
        self._qspec = self._qspec.newp(select=outsig)
        return self

    # --------------------------------------------------------------------------
    # The distinct flag
    # --------------------------------------------------------------------------
    @_generate
    @_check_join_called_first
    def distinct(self: SelfQuery) -> SelfQuery:
        self._qspec = self._qspec.newp(distinct=True)
        return self

    # --------------------------------------------------------------------------
    # Ground - bind
    # --------------------------------------------------------------------------
    @_generate
    @_check_join_called_first
    def bind(self: SelfQuery, *args: Any, **kwargs: Any) -> SelfQuery:
        self._qspec = self._qspec.bindp(*args, **kwargs)
        return self

    # --------------------------------------------------------------------------
    # The tuple flag
    # --------------------------------------------------------------------------
    @_generate
    @_check_join_called_first
    def tuple(self) -> "BaseQueryImpl[Any]":
        self._qspec = self._qspec.newp(tuple=True)
        return self

    # --------------------------------------------------------------------------
    # Overide the default heuristic
    # --------------------------------------------------------------------------
    @_generate
    def heuristic(self: SelfQuery, join_order: Any) -> SelfQuery:
        self._qspec = self._qspec.newp(heuristic=True, joh=join_order)
        return self

    # --------------------------------------------------------------------------
    # End points that do something useful
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # For the user to see what the query plan looks like
    # --------------------------------------------------------------------------
    @_check_join_called_first
    def query_plan(self, *args, **kwargs):
        qspec = self._qspec.fill_defaults()

        (factsets, factindexes) = QueryExecutor.get_factmap_data(self._factmaps, qspec)
        return make_query_plan(factindexes.keys(), qspec)

    # --------------------------------------------------------------------------
    # Return the placeholders
    # --------------------------------------------------------------------------
    @property
    def qspec(self):
        return self._qspec

    # --------------------------------------------------------------------------
    # Select to display all the output of the query
    # --------------------------------------------------------------------------
    @_check_join_called_first(endpoint=True)
    def all(self) -> Generator[_T, None, None]:
        qe = QueryExecutor(self._factmaps, self._qspec)
        return qe.all()

    # --------------------------------------------------------------------------
    # Show the single element and throw an exception if there is more than one
    # --------------------------------------------------------------------------
    @_check_join_called_first(endpoint=True)
    def singleton(self) -> _T:
        qe = QueryExecutor(self._factmaps, self._qspec)
        gen = qe.all()
        first = next(gen, None)
        if first is None:
            raise ValueError("Query has no matching elements")
        second = next(gen, None)
        if second is not None:
            raise ValueError("Query returned more than a single element")
        return first

    # --------------------------------------------------------------------------
    # Return the count of elements - Note: the behaviour of what is counted
    # changes if group_by() has been specified.
    # --------------------------------------------------------------------------
    @overload
    def count(self: "GroupedQuery[_T0, Any]") -> Iterator[Tuple[_T0, int]]: ...  # type: ignore

    @overload
    def count(self: "BaseQueryImpl[Any]") -> int: ...

    @_check_join_called_first(endpoint=True)
    def count(self) -> Union[Iterator[Tuple[Any, int]], int]:
        qe = QueryExecutor(self._factmaps, self._qspec)

        def group_by_generator():
            for k, g in qe.all():
                yield k, sum(1 for _ in g)

        # If group_by is set then we want to count records associated with each
        # key and not just total records.
        if self._qspec.group_by:
            return group_by_generator()
        else:
            return sum(1 for _ in qe.all())

    # --------------------------------------------------------------------------
    # Return the first element of the query
    # --------------------------------------------------------------------------
    @_check_join_called_first(endpoint=True)
    def first(self) -> _T:
        qe = QueryExecutor(self._factmaps, self._qspec)

        for out in qe.all():
            return out
        raise ValueError("Query has no matching elements")

    # --------------------------------------------------------------------------
    # Delete a selection of fact
    # --------------------------------------------------------------------------
    @_check_join_called_first(endpoint=True)
    def delete(self) -> int:
        qe = QueryExecutor(self._factmaps, self._qspec)
        return qe.delete()

    # --------------------------------------------------------------------------
    # Replace a selection of fact
    # --------------------------------------------------------------------------
    @_check_join_called_first(endpoint=True)
    def replace(self, fn) -> Tuple[int, int]:
        qe = QueryExecutor(self._factmaps, self._qspec)
        return qe.replace(fn)

    # --------------------------------------------------------------------------
    # Modify a selection of fact
    # --------------------------------------------------------------------------
    @_check_join_called_first(endpoint=True)
    def modify(self, fn) -> Tuple[int, int]:
        qe = QueryExecutor(self._factmaps, self._qspec)
        return qe.modify(fn)


class UnGroupedQuery(BaseQueryImpl[_T], Generic[_T]):

    # --------------------------------------------------------------------------
    # Add a group_by expression
    # --------------------------------------------------------------------------
    # START OVERLOADED FUNCTIONS self.group_by;GroupedQuery[{0}, _T];1;5;Type;Y

    # code within this block is **programmatically,
    # statically generated** by generate_overloads.py

    @overload
    def group_by(self, __ent0: Type[_T0]) -> "GroupedQuery[_T0, _T]": ...

    @overload
    def group_by(self, __ent0: _T0) -> "GroupedQuery[_T0, _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1]
    ) -> "GroupedQuery[Tuple[_T0, _T1], _T]": ...

    @overload
    def group_by(self, __ent0: Type[_T0], __ent1: _T1) -> "GroupedQuery[Tuple[_T0, _T1], _T]": ...

    @overload
    def group_by(self, __ent0: _T0, __ent1: Type[_T1]) -> "GroupedQuery[Tuple[_T0, _T1], _T]": ...

    @overload
    def group_by(self, __ent0: _T0, __ent1: _T1) -> "GroupedQuery[Tuple[_T0, _T1], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3], _T]": ...

    @overload
    def group_by(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: _T4,
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: _T3,
        __ent4: Type[_T4],
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: _T2,
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self,
        __ent0: Type[_T0],
        __ent1: _T1,
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self,
        __ent0: _T0,
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    @overload
    def group_by(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4], _T]": ...

    # END OVERLOADED FUNCTIONS self.group_by

    @overload
    def group_by(self, *expressions: Any) -> "GroupedQuery[Any, _T]": ...

    def group_by(self, *expressions: Any) -> "GroupedQuery[Any, _T]":
        if not expressions:
            nqspec = self._qspec.newp(group_by=None)  # raise exception
        else:
            nqspec = self._qspec.newp(group_by=process_orderby(expressions, self._qspec.roots))
        return GroupedQuery(self._factmaps, nqspec)

    # --------------------------------------------------------------------------
    # Explicitly select the elements to output or delete
    # --------------------------------------------------------------------------

    @overload
    def select(self, __ent0: Callable[..., _T0]) -> "UnGroupedQuery[_T0]": ...

    # START OVERLOADED FUNCTIONS self.select;UnGroupedQuery[{0}];1;5;Type;Y

    # code within this block is **programmatically,
    # statically generated** by generate_overloads.py

    @overload
    def select(self, __ent0: Type[_T0]) -> "UnGroupedQuery[_T0]": ...

    @overload
    def select(self, __ent0: _T0) -> "UnGroupedQuery[_T0]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1]]": ...

    @overload
    def select(self, __ent0: Type[_T0], __ent1: _T1) -> "UnGroupedQuery[Tuple[_T0, _T1]]": ...

    @overload
    def select(self, __ent0: _T0, __ent1: Type[_T1]) -> "UnGroupedQuery[Tuple[_T0, _T1]]": ...

    @overload
    def select(self, __ent0: _T0, __ent1: _T1) -> "UnGroupedQuery[Tuple[_T0, _T1]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: _T4,
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: _T3,
        __ent4: Type[_T4],
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: _T2,
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: _T1,
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: _T0,
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    # END OVERLOADED FUNCTIONS self.select

    @overload
    def select(self, *outsig: Any) -> "UnGroupedQuery[Any]": ...

    def select(self, *outsig: Any) -> Any:
        return super().select(*outsig)


_KT = TypeVar("_KT", bound=Any)  # Key type of GroupedQuery
_GT = TypeVar("_GT", bound=Any)  # Group type of GroupedQuery


class GroupedQuery(BaseQueryImpl[Tuple[_KT, Iterator[_GT]]], Generic[_KT, _GT]):
    def group_by(self, *expressions):
        # ABC 'Query' requires to implement group_by, but multiple group_by are not allowed
        # so we raise an error here
        # a better solution would be to somehow adjust 'Query' that GroupedQuery
        # doesn't need to implement group_by
        raise ValueError("Cannot specify 'group_by' multiple times")

    @overload
    def select(self, __ent0: Callable[..., _T0]) -> "GroupedQuery[_KT, _T0]": ...

    # START OVERLOADED FUNCTIONS self.select;GroupedQuery[_KT, {0}];1;5;Type;Y

    # code within this block is **programmatically,
    # statically generated** by generate_overloads.py

    @overload
    def select(self, __ent0: Type[_T0]) -> "GroupedQuery[_KT, _T0]": ...

    @overload
    def select(self, __ent0: _T0) -> "GroupedQuery[_KT, _T0]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1]]": ...

    @overload
    def select(self, __ent0: Type[_T0], __ent1: _T1) -> "GroupedQuery[_KT, Tuple[_T0, _T1]]": ...

    @overload
    def select(self, __ent0: _T0, __ent1: Type[_T1]) -> "GroupedQuery[_KT, Tuple[_T0, _T1]]": ...

    @overload
    def select(self, __ent0: _T0, __ent1: _T1) -> "GroupedQuery[_KT, Tuple[_T0, _T1]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: _T4,
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: _T3,
        __ent4: Type[_T4],
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: _T2,
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: Type[_T0],
        __ent1: _T1,
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: Type[_T0], __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self,
        __ent0: _T0,
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: Type[_T1], __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: Type[_T2], __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: Type[_T3], __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: Type[_T4]
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    @overload
    def select(
        self, __ent0: _T0, __ent1: _T1, __ent2: _T2, __ent3: _T3, __ent4: _T4
    ) -> "GroupedQuery[_KT, Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    # END OVERLOADED FUNCTIONS self.select

    @overload
    def select(self, *outsig: Any) -> "GroupedQuery[_KT, Any]": ...

    def select(self, *outsig: Any) -> Any:
        return super().select(*outsig)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



import importlib.util
import inspect
import secrets
import sys
import tempfile
import textwrap
import unittest
from contextlib import contextmanager
from pathlib import Path
from types import FunctionType

from clingo import Number, String

__all__ = [
    "ForwardRefTestCase",
]


def _extract_source_code_from_function(function):
    if function.__code__.co_argcount:
        raise RuntimeError(f"function {function.__qualname__} cannot have any arguments")

    code_lines = ""
    body_started = False
    for line in textwrap.dedent(inspect.getsource(function)).split("\n"):
        if line.startswith("def "):
            body_started = True
            continue
        elif body_started:
            code_lines += f"{line}\n"

    return textwrap.dedent(code_lines)


def _create_module_file(code, tmp_path, name):
    name = f"{name}_{secrets.token_hex(5)}"
    path = Path(tmp_path, f"{name}.py")
    path.write_text(code)
    return name, str(path)


def create_module(tmp_path, method_name):
    def run(source_code_or_function):
        """
        Create module object, execute it and return

        :param source_code_or_function string or function with body as a source code for created module

        """
        if isinstance(source_code_or_function, FunctionType):
            source_code = _extract_source_code_from_function(source_code_or_function)
        else:
            source_code = source_code_or_function

        module_name, filename = _create_module_file(source_code, tmp_path, method_name)

        spec = importlib.util.spec_from_file_location(module_name, filename, loader=None)
        sys.modules[module_name] = module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    return run


class ForwardRefTestCase(unittest.TestCase):
    def setUp(self):
        @contextmanager
        def f(source_code_or_function):
            with tempfile.TemporaryDirectory() as tmp_path:
                yield create_module(tmp_path, self._testMethodName)(source_code_or_function)

        self._create_module = f

    def test_postponed_annotations(self):
        code = """
from __future__ import annotations
from clorm import Predicate

class P1(Predicate):
    a: int
    b: str
"""
        with self._create_module(code) as module:
            p = module.P1(a=3, b="42")
            self.assertEqual(str(p), 'p1(3,"42")')

    def test_postponed_annotations_inner_class_basic(self):
        code = """
from __future__ import annotations
from clorm import Predicate

class Outer(Predicate):
    class Inner(Predicate):
        a: int
    b: Inner
"""
        with self._create_module(code) as module:
            inner = module.Outer.Inner(a=3)
            outer = module.Outer(b=inner)
            self.assertEqual(str(outer), "outer(inner(3))")

    def test_postponed_annotations_inner_class_complex(self):
        """Test a complex case with inner classes across different bases classes.

        Note: I think the internal implementation functionality of this test is already covered
        by other tests, but adding this for peace of mind.

        """
        code = """
from __future__ import annotations
from clorm import Predicate

class Outer(Predicate):
    class Inner(Predicate):
        a: int
    b: Inner

class NonPredicate:
    class Inner(Predicate):
        a: Outer
        b: Outer.Inner
"""
        with self._create_module(code) as module:
            inner = module.Outer.Inner(a=3)
            outer = module.Outer(b=inner)
            npinner = module.NonPredicate.Inner(a=outer, b=inner)
            self.assertEqual(str(npinner), "inner(outer(inner(3)),inner(3))")

    def test_postponed_annotations_complex(self):
        code = """
from __future__ import annotations
from clorm import Predicate
from typing import Union

class P1(Predicate):
    a: int
    b: str

class P2(Predicate):
    a: int

class P3(Predicate):
    a: 'Union[P1, P2]'
"""
        with self._create_module(code) as module:
            p = module.P3(a=module.P1(a=3, b="42"))
            self.assertEqual(str(p), 'p3(p1(3,"42"))')
            p = module.P3(a=module.P2(a=42))
            self.assertEqual(str(p), "p3(p2(42))")

    def test_postponed_annotations_complex2(self):
        code = """
from __future__ import annotations
from clorm import Predicate
from typing import Union

class P1(Predicate):
    a: int
    b: str

class P2(Predicate):
    a: int

class P3(Predicate):
    a: P1 | P2
"""
        if sys.version_info >= (3, 10):
            with self._create_module(code) as module:
                p = module.P3(a=module.P1(a=3, b="42"))
                self.assertEqual(str(p), 'p3(p1(3,"42"))')
                p = module.P3(a=module.P2(a=42))
                self.assertEqual(str(p), "p3(p2(42))")

    def test_postponed_annotations_complex3(self):
        code = """
from __future__ import annotations
from clorm import Predicate
from typing import Union

class P1(Predicate):
    a: int
    b: str

class P2(Predicate):
    a: int

class P3(Predicate):
    a: 'P1 | P2'
"""
        if sys.version_info >= (3, 10):
            with self._create_module(code) as module:
                p = module.P3(a=module.P1(a=3, b="42"))
                self.assertEqual(str(p), 'p3(p1(3,"42"))')
                p = module.P3(a=module.P2(a=42))
                self.assertEqual(str(p), "p3(p2(42))")

    def test_postponed_annotations_tuple1(self):
        code = """
from __future__ import annotations
from clorm import Predicate

class P(Predicate):
    a: tuple[int, str]
"""
        if sys.version_info >= (3, 10):
            with self._create_module(code) as module:
                p = module.P(a=(123, "Hello"))
                self.assertEqual(str(p), 'p((123,"Hello"))')

    def test_postponed_annotations_tuple1(self):
        code = """
from __future__ import annotations
from clorm import Predicate
from typing import Tuple

class P(Predicate):
    a: Tuple[int, str]
"""
        with self._create_module(code) as module:
            p = module.P(a=(123, "Hello"))
            self.assertEqual(str(p), 'p((123,"Hello"))')

    def test_postponed_annotations_nonglobal1(self):
        code = """
from __future__ import annotations
from clorm import Predicate, ConstantField, field
from typing import Union

def define_predicates():


    class P1(Predicate):
        a1: str = field(ConstantField)
        a: int
        b: str

    class P2(Predicate):
        a: Union[int, P1]

    return P1, P2

XP1, XP2 = define_predicates()

"""
        with self._create_module(code) as module:
            p1 = module.XP1(a1="c", a=3, b="42")
            self.assertEqual(str(p1), 'p1(c,3,"42")')
            p2 = module.XP2(a=p1)
            self.assertEqual(str(p2), 'p2(p1(c,3,"42"))')

    def test_postponed_annotations_nonglobal2(self):
        code = """
from __future__ import annotations
from clorm import Predicate, ConstantField, field
from typing import Union

def define_predicates():


    class P1(Predicate):
        a1: str = field(ConstantField)
        a: int
        b: str

    def define_complex():
        class P2(Predicate):
            a: Union[int, P1]
        return P2

    return P1, define_complex()

XP1, XP2 = define_predicates()

"""
        with self._create_module(code) as module:
            p1 = module.XP1(a1="c", a=3, b="42")
            self.assertEqual(str(p1), 'p1(c,3,"42")')
            p2 = module.XP2(a=p1)
            self.assertEqual(str(p2), 'p2(p1(c,3,"42"))')

    def test_postponed_annotations_headlist(self):
        code = """
from __future__ import annotations
from typing import Tuple
from clorm import Predicate, HeadList

class P(Predicate):
    x: HeadList[Tuple[int,str]]
"""
        with self._create_module(code) as module:
            p = module.P(x=((1, "a"), (2, "b")))
            self.assertEqual(str(p), 'p(((1,"a"),((2,"b"),())))')

    def test_postponed_annotations_flatlist(self):
        code = """
from __future__ import annotations
from typing import Tuple
from clorm import Predicate

class P(Predicate):
    x: Tuple[Tuple[int,str], ...]
"""
        with self._create_module(code) as module:
            p = module.P(x=((1, "a"), (2, "b")))
            self.assertEqual(str(p), 'p(((1,"a"),(2,"b")))')

    def test_forward_ref(self):
        def module_():
            from typing import ForwardRef

            from clorm import Predicate

            class A(Predicate):
                a: int

            ARef = ForwardRef("A")

            class B(Predicate):
                a: ARef

        with self._create_module(module_) as module:
            b = module.B(a=module.A(a=42))
            self.assertEqual(str(b), "b(a(42))")

    def test_forward_ref_list(self):
        def module_():
            from typing import ForwardRef

            from clorm import HeadList, Predicate

            class A(Predicate):
                a: int

            ARef = ForwardRef("A")

            class B(Predicate):
                a: HeadList[ARef]

        with self._create_module(module_) as module:
            b = module.B(a=[module.A(a=41), module.A(a=42)])
            self.assertEqual(str(b), "b((a(41),(a(42),())))")

    def test_forward_ref_asp_callable(self):
        code = """
from __future__ import annotations
from clorm import Predicate, make_function_asp_callable, make_method_asp_callable

class P1(Predicate):
    a: int
    b: str

@make_function_asp_callable
def f(a: int, b: str) -> P1:
    return P1(a,b)

class Context:
    @make_method_asp_callable
    def f(self, a: int, b: str) -> P1:
        return P1(a,b)
"""
        with self._create_module(code) as module:
            p = module.f(Number(2), String("2"))
            self.assertEqual(str(p), 'p1(2,"2")')
            ctx = module.Context()
            p = ctx.f(Number(2), String("2"))
            self.assertEqual(str(p), 'p1(2,"2")')



# ------------------------------------------------------------------------------
# Unit tests for the bool library
# ------------------------------------------------------------------------------


import unittest

from clorm.lib.boolean import BooleanField, StrictBooleanField
from clorm.orm.noclingo import Number, String
from tests.support import check_errmsg


class LibBoolTestCase(unittest.TestCase):

    # --------------------------------------------------------------------------
    # Make sure BooleanField does what we expect
    # --------------------------------------------------------------------------
    def test_booleanfield(self):
        symstr = Number(0)
        self.assertEqual(type(BooleanField.cltopy(symstr)), bool)
        self.assertEqual(BooleanField.cltopy(symstr), False)
        self.assertEqual(BooleanField.pytocl(0), symstr)

        symstr = String("false")
        self.assertEqual(type(BooleanField.cltopy(symstr)), bool)
        self.assertEqual(BooleanField.cltopy(symstr), False)

        self.assertEqual(BooleanField.pytocl("false"), Number(0))
        self.assertEqual(BooleanField.pytocl("TrUe"), Number(1))
        self.assertEqual(BooleanField.pytocl("on"), Number(1))
        self.assertEqual(BooleanField.pytocl("off"), Number(0))

        with self.assertRaises(TypeError) as ctx:
            BooleanField.cltopy(String("2"))
        check_errmsg("value '2'", ctx)

    # --------------------------------------------------------------------------
    # Make sure StrictBooleanField does what we expect
    # --------------------------------------------------------------------------
    def test_strictbooleanfield(self):
        symstr = Number(1)
        self.assertEqual(type(StrictBooleanField.cltopy(symstr)), bool)
        self.assertEqual(StrictBooleanField.cltopy(symstr), True)
        self.assertEqual(StrictBooleanField.pytocl(True), symstr)

        with self.assertRaises(TypeError) as ctx:
            StrictBooleanField.cltopy(Number(2))
        check_errmsg("value must be either", ctx)



"""Predicate sub-class function templates.

Some of the functions of a Predicate sub-class are generated dynamically using
an `exec()` call. This file contains some template code snippets and functions
to help apply these templates.

"""

from __future__ import annotations

# ------------------------------------------------------------------------------
# Helper functions for PredicateMeta class to create a Predicate
# class constructor.
# ------------------------------------------------------------------------------


def expand_template(template: str, **kwargs: str) -> str:
    """Expand the template by substituting the arguments.

    If the argument contains multiple lines then extra spaces are added to each
    line to preserve the correct indentation.

    """

    # Add spaces to each line of some multi-text input
    def add_spaces(num, text):
        space = " " * num
        out = []
        for idx, line in enumerate(text.splitlines()):
            if idx == 0:
                out = [line]
            else:
                out.append(space + line)
        return "\n".join(out)

    if not template:
        return ""
    lines = template.expandtabs(4).splitlines()
    outlines = []
    for line in lines:
        start = line.find(r"{%")
        if start == -1:
            outlines.append(line)
            continue
        end = line.find(r"%}", start)
        if end == -1:
            raise ValueError("Bad template expansion in {line}")
        keyword = line[start + 2 : end]
        text = add_spaces(start, kwargs[keyword])
        line = line[0:start] + text + line[end + 2 :]
        outlines.append(line)
    return "\n".join(outlines)


PREDICATE_TEMPLATE = r"""
def __init__(self,
             {{%args_signature%}}
             *, sign: bool=True) -> None:

    self._hash = None
    self._sign = bool(sign)

    {{%sign_check%}}
    {{%check_no_defaults%}}

    # Assign defaults for missing values and apply map tuple transform for complex values
    {{%assign_defaults%}}
    {{%check_complex%}}

    self._field_values = ({{%args%}})

    # Create the raw symbol
    self._raw = Function("{pdefn.name}",
                         ({{%args_raw%}}),
                         self._sign)


@classmethod
def _unify(cls: Type[_P], raw: AnySymbol, raw_args: Optional[Sequence[AnySymbol]]=None, raw_name: Optional[str]=None) -> Optional[_P]:
    try:
        raw_args = raw_args if raw_args else raw.arguments
        raw_name = raw_name if raw_name else raw.name
        if len(raw_args) != {pdefn.arity}:
            return None

        {{%sign_check_unify%}}

        if raw_name != "{pdefn.name}":
            return None

        instance = cls.__new__(cls)
        instance._raw = raw
        instance._hash = None
        instance._sign = raw.positive
        instance._field_values = ({{%args_cltopy%}})
        return instance
    except (TypeError, ValueError):
        return None
    except AttributeError as e:
        raise ValueError((f"Cannot unify with object {{raw}} ({{type(raw)}}) as "
                          "it is not a clingo Symbol Function object"))


def nontuple__eq__(self, other: Any) -> bool:
    # Deal with a non-tuple predicate
    if isinstance(other, Predicate):
        return self._raw == other._raw
    if isinstance(other, Symbol):
        return self._raw == other
    return NotImplemented


def tuple__eq__(self, other: Any) -> bool:
    # Deal with a predicate that is a tuple
    if isinstance(other, Predicate):
        return self._raw == other._raw
    if isinstance(other, Symbol):
        return self._raw == other
#    if isinstance(other, tuple):
#        return self._field_values == other
    return NotImplemented


def nontuple__lt__(self, other):
    # If it is the same predicate class then compare the underlying clingo symbol
    if isinstance(other, Predicate):
        return self._raw < other._raw
    if isinstance(other, Symbol):
        return self._raw < other
    return NotImplemented


def tuple__lt__(self, other):
    # self is always less than a non-tuple predicate
    if isinstance(other, Predicate):
        return self._raw < other._raw
    if isinstance(other, Symbol):
        return self._raw < other
#    if isinstance(other, tuple):
#        return self._field_values < other
    return NotImplemented


def nontuple__gt__(self, other):
    if isinstance(other, Predicate):
        return self._raw > other._raw
    if isinstance(other, Symbol):
        return self._raw > other
    return NotImplemented


def tuple__gt__(self, other):
    # If it is the same predicate class then compare the sign and fields
    if isinstance(other, Predicate):
        return self._raw > other._raw
    if isinstance(other, Symbol):
        return self._raw > other
#    if isinstance(other, tuple):
#        return self._field_values > other
    return NotImplemented


def __ge__(self, other):
    result = self.__lt__(other)
    if result is NotImplemented:
        return NotImplemented
    return not result


def __le__(self, other):
    result = self.__gt__(other)
    if result is NotImplemented:
        return NotImplemented
    return not result


def __hash__(self):
    if self._hash is None:
        self._hash = hash(self._raw)
    return self._hash


__eq__ = tuple__eq__ if PREDICATE_IS_TUPLE else nontuple__eq__
__lt__ = tuple__lt__ if PREDICATE_IS_TUPLE else nontuple__lt__
__gt__ = tuple__gt__ if PREDICATE_IS_TUPLE else nontuple__gt__



"""

CHECK_SIGN_TEMPLATE = r"""
# Check if the sign is allowed
if self._sign != {sign}:
    raise ValueError(f"Predicate {{type(self).__name__}}"
                     f"is defined to only allow {pdefn.sign} instances")
"""

CHECK_SIGN_UNIFY_TEMPLATE = r"""
if raw.positive != {sign}:
    return None
"""

NO_DEFAULTS_TEMPLATE = r"""
# Check for missing values that have no defaults
if MISSING in ({args}):
    for arg, name in ({named_args}):
        if arg is MISSING:
            raise TypeError((f"Missing argument for field \"{{name}}\""
                             f"(which has no default value)"))
"""

ASSIGN_DEFAULT_TEMPLATE = r"""
if {arg} is MISSING:
    {arg} = {arg}_field.default
"""

ASSIGN_COMPLEX_TEMPLATE = r"""
if not isinstance({arg}, {arg}_class):
    if isinstance({arg}, tuple) or (isinstance({arg}, Predicate) and {arg}.meta.is_tuple):
        {arg} = {arg}_class(*{arg})
    else:
        raise TypeError(f"Value {{{arg}}} ({{type({arg})}}) is not a tuple")
"""

PREDICATE_UNIFY_DOCSTRING = r"""
    Unify a (raw) Symbol object with the class.

    Returns None on failure to unify otherwise returns the new fact
"""

# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



# -----------------------------------------------------------------------------
# Clorm ORM FactBase implementation. FactBase provides a set-like container
# specifically for storing facts (Predicate instances).
# ------------------------------------------------------------------------------

from __future__ import annotations

import abc
import io
import itertools
import sys
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    TextIO,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from ._queryimpl import UnGroupedQuery
from ._typing import _T0, _T1, _T2, _T3, _T4
from .core import Predicate, PredicateDefn, PredicatePath, and_, validate_root_paths
from .factcontainers import FactMap, factset_equality
from .query import QueryExecutor, QuerySpec, make_query_plan, process_orderby, process_where

__all__ = [
    "FactBase",
    "Select",
    "Delete",
]

# ------------------------------------------------------------------------------
# Global
# ------------------------------------------------------------------------------
_Facts = Union[Iterable[Predicate], Callable[[], Iterable[Predicate]]]

# ------------------------------------------------------------------------------
# Support function for printing ASP facts: Note: _trim_docstring() is taken from
# PEP 257 (modified for Python 3): https://www.python.org/dev/peps/pep-0257/
# ------------------------------------------------------------------------------

_builtin_sorted = sorted


def _format_asp_facts(iterator, output, width):
    tmp1 = ""
    for f in iterator:
        fstr = "{}.".format(f)
        if tmp1 and len(tmp1) + len(fstr) > width:
            print(tmp1, file=output)
            tmp1 = fstr
        else:
            tmp1 = tmp1 + " " + fstr if tmp1 else fstr
    if tmp1:
        print(tmp1, file=output)


def _trim_docstring(docstring):
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return "\n".join(trimmed)


def _endstrip(string):
    if not string:
        return
    nl = string[-1] == "\n"
    tmp = string.rstrip()
    return tmp + "\n" if nl else tmp


def _format_docstring(docstring, output):
    if not docstring:
        return
    tmp = _trim_docstring(docstring)
    tmpstr = "".join(_endstrip("%     " + l) for l in tmp.splitlines(True))
    if tmpstr:
        print("% Description:", file=output)
        print(tmpstr, file=output)


def _maxwidth(lines):
    return max([len(l) for l in lines])


def _format_commented(fm: FactMap, out: TextIO) -> None:
    pm: PredicateDefn = fm.predicate.meta
    docstring = _trim_docstring(fm.predicate.__doc__) if fm.predicate.__doc__ else ""
    indent = "    "
    if pm.arity == 0:
        lines = ["Unary predicate signature:", indent + pm.name]
    else:

        def build_signature(p: Type[Predicate]) -> str:
            args = []
            for pp in p:
                complex = pp.meta.field.complex
                args.append(
                    cast(str, pp._pathseq[1]) if not complex else build_signature(complex)
                )
            return f"{p.meta.name}({','.join(args)})"

        lines = ["Predicate signature:", indent + build_signature(fm.predicate)]
    if docstring:
        lines.append("Description:")
        for l in docstring.splitlines():
            lines.append(indent + l)
    bar = "-" * _maxwidth(lines)
    lines.insert(0, bar)
    lines.append(bar)
    for l in lines:
        tmp = l.rstrip()
        if tmp:
            print("% {}".format(tmp), file=out)
        else:
            print("%", file=out)
    return


# ------------------------------------------------------------------------------
# A FactBase consisting of facts of different types
# ------------------------------------------------------------------------------


class FactBase(object):
    """A fact base is a container for facts (i.e., Predicate sub-class instances)

    ``FactBase`` can be behave like a specialised ``set`` object, but can also
    behave like a minimalist database. It stores facts for ``Predicate`` types
    (where a predicate type loosely corresponds to a *table* in a database)
    and allows for certain fields to be indexed in order to perform more
    efficient queries.

    The initaliser can be given a collection of predicates. If it is passed
    another FactBase then it simply makes a copy (including the indexed fields).

    FactBase also has a special mode when it is passed a functor instead of a
    collection. In this case it performs a delayed initialisation. This means
    that the internal data structures are only populated when the FactBase is
    actually used. This mode is particularly useful when extracting facts from
    models. Often a program will only want to keep the data from the final model
    (for example, with optimisation we often want the best model before a
    timeout). Delayed initialisation is useful will save computation as only the
    last model will be properly initialised.

    Args:
      facts([Predicate]|FactBase|callable): a list of facts (predicate
         instances), a fact base, or a functor that generates a list of
         facts. If a functor is passed then the fact base performs a delayed
         initialisation. If a fact base is passed and no index is specified then
         an index will be created matching in input fact base.
      indexes(Field): a list of fields that are to be indexed.

    """

    # --------------------------------------------------------------------------
    # Internal member functions
    # --------------------------------------------------------------------------

    # A special purpose initialiser so that we can delayed initialisation
    def _init(self, facts=None, indexes=None):

        # flag that initialisation has taken place
        self._delayed_init: Optional[Callable[[], None]] = None

        # If it is delayed initialisation then get the facts
        if facts and callable(facts):
            facts = facts()
        elif facts and isinstance(facts, FactBase) and indexes is None:
            indexes = facts.indexes
        if indexes is None:
            indexes = []

        # Create FactMaps for the predicate types with indexed fields
        grouped = {}

        self._indexes = tuple(indexes)
        for path in self._indexes:
            if path.meta.predicate not in grouped:
                grouped[path.meta.predicate] = []
            grouped[path.meta.predicate].append(path)
        self._factmaps = {pt: FactMap(pt, idxs) for pt, idxs in grouped.items()}

        if facts is None:
            return
        self._add(facts)

    # Make sure the FactBase has been initialised
    def _check_init(self):
        if self._delayed_init:
            self._delayed_init()  # Check for delayed init

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------

    def _add(self, arg: Union[Predicate, Iterable[Predicate]]) -> None:
        if isinstance(arg, Predicate):
            ptype = arg.__class__
            if not ptype in self._factmaps:
                self._factmaps[ptype] = FactMap(ptype)
            return self._factmaps[ptype].add_fact(arg)

        if isinstance(arg, str) or not isinstance(arg, Iterable):
            raise TypeError(f"'{arg}' is not a Predicate instance")

        sorted_facts = sorted(arg, key=lambda x: x.__class__.__name__)
        for type_, grouped_facts in itertools.groupby(sorted_facts, lambda x: x.__class__):
            if not issubclass(type_, Predicate):
                raise TypeError(f"{list(grouped_facts)} are not Predicate instances")
            if not type_ in self._factmaps:
                self._factmaps[type_] = FactMap(type_)
            self._factmaps[type_].add_facts(grouped_facts)
        return

    def _remove(self, fact, raise_on_missing):
        ptype = type(fact)
        if not isinstance(fact, Predicate) or ptype not in self._factmaps:
            raise KeyError(fact)

        return self._factmaps[ptype].remove(fact, raise_on_missing)

    # --------------------------------------------------------------------------
    # Initiliser
    # --------------------------------------------------------------------------
    def __init__(
        self, facts: Optional[_Facts] = None, indexes: Optional[Iterable[PredicatePath]] = None
    ) -> None:
        self._delayed_init = None
        if callable(facts):

            def delayed_init():
                self._init(facts, indexes)

            self._delayed_init = delayed_init
        else:
            self._init(facts, indexes)

    # --------------------------------------------------------------------------
    # An internal API for the query mechanism. Not to be called by users.
    # --------------------------------------------------------------------------
    @property
    def factmaps(self):
        self._check_init()  # Check for delayed init
        return self._factmaps

    # --------------------------------------------------------------------------
    # Set member functions
    # --------------------------------------------------------------------------
    def add(self, arg: Union[Predicate, Iterable[Predicate]]) -> None:
        """Add a single fact or a collection of facts.

        Because a ``FactBase`` can only hold :class:`~Predicate` sub-class
        instances this member function has been overloaded to take either a
        single :class:`~Predicate` sub-class instance or a collection of
        :class:`~Predicate` sub-class instances.

        Args:
          arg: a single fact or a collection of facts.

        """
        self._check_init()  # Check for delayed init
        return self._add(arg)

    def remove(self, arg: Predicate) -> None:
        """Remove a fact from the fact base (raises an exception if no fact)."""
        self._check_init()  # Check for delayed init
        return self._remove(arg, raise_on_missing=True)

    def discard(self, arg: Predicate) -> None:
        """Remove a fact from the fact base."""
        self._check_init()  # Check for delayed init
        return self._remove(arg, raise_on_missing=False)

    def pop(self) -> Predicate:
        """Pop an element from the FactBase."""
        self._check_init()  # Check for delayed init
        for pt, fm in self._factmaps.items():
            if fm:
                return fm.pop()
        raise KeyError("pop from an empty FactBase")

    def clear(self):
        """Clear the fact base of all facts."""

        self._check_init()  # Check for delayed init
        for pt, fm in self._factmaps.items():
            fm.clear()

    # --------------------------------------------------------------------------
    # Special FactBase member functions
    # --------------------------------------------------------------------------
    def select(self, root):
        """Define a select query using the old Query API.

        .. note::

           This interface will eventually be deprecated when the new
           :class:`Query API<Query>` is finalised. The entry point to this Query
           API is through the :meth:`FactBase.query` method.

        Args:
           predicate: The predicate to query.

        Returns:
           Returns a Select query object for specifying a query.

        """
        import warnings

        warnings.warn(
            "This API is deprecated. To use the new Query API call FactBase.query()",
            DeprecationWarning,
            stacklevel=2,
        )
        self._check_init()  # Check for delayed init

        roots = validate_root_paths([root])
        ptypes = set([root.meta.predicate for root in roots])

        # Make sure there are factmaps for each referenced predicate type
        for ptype in ptypes:
            self._factmaps.setdefault(ptype, FactMap(ptype))

        return SelectImpl(self, QuerySpec(roots=roots))

    def delete(self, root):
        self._check_init()  # Check for delayed init

        roots = validate_root_paths([root])
        ptypes = set([root.meta.predicate for root in roots])

        # Make sure there are factmaps for each referenced predicate type
        for ptype in ptypes:
            self._factmaps.setdefault(ptype, FactMap(ptype))

        return _Delete(self, QuerySpec(roots=roots))

    # START OVERLOADED FUNCTIONS self.query;UnGroupedQuery[{0}];1;5;Type;

    # code within this block is **programmatically,
    # statically generated** by generate_overloads.py

    @overload
    def query(self, __ent0: Type[_T0]) -> "UnGroupedQuery[_T0]": ...

    @overload
    def query(
        self, __ent0: Type[_T0], __ent1: Type[_T1]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1]]": ...

    @overload
    def query(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2]]": ...

    @overload
    def query(
        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]": ...

    @overload
    def query(
        self,
        __ent0: Type[_T0],
        __ent1: Type[_T1],
        __ent2: Type[_T2],
        __ent3: Type[_T3],
        __ent4: Type[_T4],
    ) -> "UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]": ...

    # END OVERLOADED FUNCTIONS self.query

    @overload
    def query(self, *roots: Any) -> "UnGroupedQuery[Any]": ...

    def query(self, *roots):
        """Define a query using the new Query API :class:`Query`.

        The parameters consist of a predicates (or aliases) to query (like an
        SQL FROM clause).

        Args:
           *predicates: predicate or predicate aliases

        Returns:
           Returns a Query object for specifying a query.

        """

        self._check_init()  # Check for delayed init

        # Make sure there are factmaps for each referenced predicate type
        ptypes = set([r.meta.predicate for r in validate_root_paths(roots)])
        for ptype in ptypes:
            self._factmaps.setdefault(ptype, FactMap(ptype))

        qspec = QuerySpec(roots=roots)
        return UnGroupedQuery(self._factmaps, qspec)

    @property
    def predicates(self) -> Tuple[Type[Predicate], ...]:
        """Return the list of predicate types that this fact base contains."""

        self._check_init()  # Check for delayed init
        return tuple([pt for pt, fm in self._factmaps.items() if fm])

    @property
    def indexes(self) -> Tuple[PredicatePath, ...]:
        self._check_init()  # Check for delayed init
        return self._indexes

    def facts(self) -> List[Predicate]:
        """Return all facts."""

        self._check_init()  # Check for delayed init
        tmp = [fm.factset for fm in self._factmaps.values() if fm]
        return list(itertools.chain(*tmp))

    def asp_str(self, *, width: int = 0, commented: bool = False, sorted: bool = False) -> str:
        """Return a ASP string representation of the fact base.

        The generated ASP string representation is syntactically correct ASP
        code so is suitable for adding as the input to to an ASP program (or
        writing to a file for later use in an ASP program).

        By default the order of the facts in the string is arbitrary. Because
        `FactBase` is built on a `OrderedDict` (which preserves insertion
        order) the order of the facts will be deterministic between runs of the
        same program. However two FactBases containing the same facts but
        constructed in different ways will not produce the same output
        string. In order to guarantee the same output the `sorted` flag can be
        specified.

        Args:
            width: tries to fill to a given width by putting more than one
                   fact on a line if necessary (default: 0).
            commented: produces commented ASP code by adding a predicate
                       signature and turning the Predicate sub-class docstring
                       into a ASP comments (default: False).
            sorted: sort the output facts, first by predicates (name,arity) and
                    then by the natural order of the instances for that
                    predicate (default :False).

        """
        self._check_init()  # Check for delayed init
        out = io.StringIO()

        first = True
        if sorted:
            names = _builtin_sorted(
                self._factmaps.keys(), key=lambda pt: (pt.meta.name, pt.meta.arity, pt.__name__)
            )
            fms = [self._factmaps[n] for n in names]
        else:
            fms = self._factmaps.values()
        for fm in fms:
            if commented:
                if first:
                    first = False
                else:
                    print("", file=out)
                _format_commented(fm, out)
            if sorted:
                _format_asp_facts(_builtin_sorted(fm.factset), out, width)
            else:
                _format_asp_facts(fm.factset, out, width)

        data = out.getvalue()
        out.close()
        return data

    def __str__(self) -> str:
        self._check_init()  # Check for delayed init
        tmp = ", ".join([str(f) for f in self])
        return "{" + tmp + "}"

    def __repr__(self):
        return self.__str__()

    # --------------------------------------------------------------------------
    # Special functions to support set and container operations
    # --------------------------------------------------------------------------

    def __contains__(self, fact):
        """Implemement set 'in' operator."""

        self._check_init()  # Check for delayed init

        if not isinstance(fact, Predicate):
            return False
        ptype = type(fact)
        if ptype not in self._factmaps:
            return False
        return fact in self._factmaps[ptype].factset

    def __bool__(self):
        """Implemement set bool operator."""

        self._check_init()  # Check for delayed init

        for fm in self._factmaps.values():
            if fm:
                return True
        return False

    def __len__(self):
        self._check_init()  # Check for delayed init
        return sum([len(fm.factset) for fm in self._factmaps.values()])

    def __iter__(self) -> Iterator[Predicate]:
        self._check_init()  # Check for delayed init

        for fm in self._factmaps.values():
            for f in fm.factset:
                yield f

    def __eq__(self, other):
        """Overloaded boolean operator."""

        # If other is not a FactBase then create one
        if not isinstance(other, self.__class__):
            other = FactBase(other)
        self._check_init()
        other._check_init()  # Check for delayed init

        self_fms = {p: fm for p, fm in self._factmaps.items() if fm}
        other_fms = {p: fm for p, fm in other._factmaps.items() if fm}
        if self_fms.keys() != other_fms.keys():
            return False

        for p, fm1 in self_fms.items():
            fm2 = other_fms[p]
            if not factset_equality(fm1.factset, fm2.factset):
                return False

        return True

    def __lt__(self, other):
        """Implemement set < operator."""

        # If other is not a FactBase then create one
        if not isinstance(other, self.__class__):
            other = FactBase(other)
        self._check_init()
        other._check_init()  # Check for delayed init

        self_fms = {p: fm for p, fm in self._factmaps.items() if fm}
        other_fms = {p: fm for p, fm in other._factmaps.items() if fm}
        if len(self_fms) > len(other_fms):
            return False

        known_ne = False
        for p, spfm in self_fms.items():
            if p not in other_fms:
                return False
            opfm = other_fms[p]
            if spfm.factset < opfm.factset:
                known_ne = True
            elif spfm.factset > opfm.factset:
                return False

        if known_ne:
            return True
        return False

    def __le__(self, other):
        """Implemement set <= operator."""

        if not isinstance(other, self.__class__):
            other = FactBase(other)
        self._check_init()
        other._check_init()  # Check for delayed init

        self_fms = {p: fm for p, fm in self._factmaps.items() if fm}
        other_fms = {p: fm for p, fm in other._factmaps.items() if fm}
        if len(self_fms) > len(other_fms):
            return False

        for p, spfm in self_fms.items():
            if p not in other_fms:
                return False
            opfm = other_fms[p]
            if spfm.factset > opfm.factset:
                return False
        return True

    def __gt__(self, other):
        """Implemement set > operator."""
        if not isinstance(other, self.__class__):
            other = FactBase(other)
        return other.__lt__(self)

    def __ge__(self, other):
        """Implemement set >= operator."""
        if not isinstance(other, self.__class__):
            other = FactBase(other)
        return other.__le__(self)

    def __or__(self, other):
        """Implemement set | operator."""
        return self.union(other)

    def __and__(self, other):
        """Implemement set & operator."""
        return self.intersection(other)

    def __sub__(self, other):
        """Implemement set - operator."""
        return self.difference(other)

    def __xor__(self, other):
        """Implemement set ^ operator."""
        return self.symmetric_difference(other)

    def __ior__(self, other):
        """Implemement set |= operator."""
        self.update(other)
        return self

    def __iand__(self, other):
        """Implemement set &= operator."""
        self.intersection_update(other)
        return self

    def __isub__(self, other):
        """Implemement set -= operator."""
        self.difference_update(other)
        return self

    def __ixor__(self, other):
        """Implemement set ^= operator."""
        self.symmetric_difference_update(other)
        return self

    def __getstate__(self):
        self._check_init()
        return self.__dict__

    # --------------------------------------------------------------------------
    # Set functions
    # --------------------------------------------------------------------------
    def union(self, *others: _Facts) -> "FactBase":
        """Implements the set union() function"""
        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init()  # Check for delayed init
        for fb in factbases:
            fb._check_init()

        fb = FactBase()
        predicates = set(self._factmaps.keys())
        for o in factbases:
            predicates.update(o._factmaps.keys())

        for p in predicates:
            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
            if p in self._factmaps:
                fb._factmaps[p] = self._factmaps[p].union(*pothers)
            else:
                fb._factmaps[p] = FactMap(p).union(*pothers)
        return fb

    def intersection(self, *others: _Facts) -> "FactBase":
        """Implements the set intersection() function"""
        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init()  # Check for delayed init
        for fb in factbases:
            fb._check_init()

        fb = FactBase()
        predicates = set(self._factmaps.keys())
        for fb_ in factbases:
            predicates.intersection_update(fb_._factmaps.keys())

        for p in predicates:
            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
            fb._factmaps[p] = self._factmaps[p].intersection(*pothers)
        return fb

    def difference(self, *others: _Facts) -> "FactBase":
        """Implements the set difference() function"""
        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init()  # Check for delayed init
        for fb in factbases:
            fb._check_init()

        fb = FactBase()
        predicates = set(self._factmaps.keys())

        for p in predicates:
            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
            fb._factmaps[p] = self._factmaps[p].difference(*pothers)
        return fb

    def symmetric_difference(self, other: _Facts) -> "FactBase":
        """Implements the set symmetric_difference() function"""
        if not isinstance(other, self.__class__):
            other = FactBase(other)
        self._check_init()  # Check for delayed init
        other._check_init()

        fb = FactBase()
        predicates = set(self._factmaps.keys())
        predicates.update(other._factmaps.keys())

        for p in predicates:
            in_self = p in self._factmaps
            in_other = p in other._factmaps
            if in_self and in_other:
                fb._factmaps[p] = self._factmaps[p].symmetric_difference(other._factmaps[p])
            elif in_self:
                fb._factmaps[p] = self._factmaps[p].copy()
            elif in_other:
                fb._factmaps[p] = other._factmaps[p].copy()

        return fb

    def update(self, *others: _Facts) -> None:
        """Implements the set update() function"""
        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init()  # Check for delayed init
        for fb in factbases:
            fb._check_init()

        for fb in factbases:
            for p, fm in fb._factmaps.items():
                if p in self._factmaps:
                    self._factmaps[p].update(fm)
                else:
                    self._factmaps[p] = fm.copy()

    def intersection_update(self, *others: _Facts) -> None:
        """Implements the set intersection_update() function"""
        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init()  # Check for delayed init
        for fb in factbases:
            fb._check_init()

        predicates = set(self._factmaps.keys())
        for fb in factbases:
            predicates.intersection_update(fb._factmaps.keys())
        pred_to_delete = set(self._factmaps.keys()) - predicates

        for p in pred_to_delete:
            self._factmaps[p].clear()
        for p in predicates:
            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
            self._factmaps[p].intersection_update(*pothers)

    def difference_update(self, *others: _Facts) -> None:
        """Implements the set difference_update() function"""
        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init()  # Check for delayed init
        for fb in factbases:
            fb._check_init()

        for p in self._factmaps.keys():
            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
            self._factmaps[p].difference_update(*pothers)

    def symmetric_difference_update(self, other: _Facts) -> None:
        """Implements the set symmetric_difference_update() function"""
        if not isinstance(other, self.__class__):
            other = FactBase(other)
        self._check_init()  # Check for delayed init
        other._check_init()

        predicates = set(self._factmaps.keys())
        predicates.update(other._factmaps.keys())

        for p in predicates:
            if p in self._factmaps and p in other._factmaps:
                self._factmaps[p].symmetric_difference_update(other._factmaps[p])
            else:
                if p in other._factmaps:
                    self._factmaps[p] = other._factmaps[p].copy()

    def copy(self) -> "FactBase":
        """Implements the set copy() function"""
        self._check_init()  # Check for delayed init
        fb = FactBase()
        for p, _ in self._factmaps.items():
            fb._factmaps[p] = self._factmaps[p].copy()
        return fb


# ------------------------------------------------------------------------------
# Select is an interface query over a FactBase.
# ------------------------------------------------------------------------------


class Select(abc.ABC):
    """An abstract class that defines the interface to original Query API.

    .. note::

       This interface will eventually be deprecated when the new :class:`Query
       API<Query>` is finalised.

    ``Select`` query objects cannot be constructed directly. Instead a
    ``Select`` object is returned by the :meth:`FactBase.select` function. Given
    a ``FactBase`` object ``fb``, a specification is of the form:

          ``query = fb.select(<predicate>).where(<expression>).order_by(<ordering>)``

    where ``<predicate>`` specifies the predicate type to search for,
    ``<expression>`` specifies the search criteria and ``<ordering>`` specifies
    a sort order when returning the results. The ``where()`` and ``order_by()``
    clauses are omitted when not required.

    """

    @abc.abstractmethod
    def where(self, *expressions):
        """Set the select statement's where clause.

         The where clause consists of a set of boolean and comparison
         expressions. This expression specifies a search criteria for matching
         facts within the corresponding ``FactBase``.

         Boolean expression are built from other boolean expression or a
         comparison expression. Comparison expressions are of the form:

                ``<PredicatePath> <compop>  <value>``

        where ``<compop>`` is a comparison operator such as ``==``, ``!=``, or
        ``<=`` and ``<value>`` is either a Python value or another predicate path
        object refering to a field of the same predicate or a placeholder.

         A placeholder is a special value that issubstituted when the query is
         actually executed. These placeholders are named ``ph1_``, ``ph2_``,
         ``ph3_``, and ``ph4_`` and correspond to the 1st to 4th arguments of the
         ``get``, ``get_unique`` or ``count`` function call.

         Args:
           expressions: one or more comparison expressions.

         Returns:
           Returns a reference to itself.

        """

    @abc.abstractmethod
    def order_by(self, *fieldorder):
        """Provide an ordering over the results.

        Args:
          fieldorder: an ordering over fields
        Returns:
          Returns a reference to itself.
        """
        pass

    @abc.abstractmethod
    def get(self, *args, **kwargs):
        """Return all matching entries."""
        pass

    def get_unique(self, *args, **kwargs):
        """Return the unique matching entry (or raise an exception)"""
        pass

    def count(self, *args, **kwargs):
        """Return the number of matches."""
        pass


# ------------------------------------------------------------------------------
# Delete is an interface to perform a query delete from a FactBase.
# ------------------------------------------------------------------------------


class Delete(abc.ABC):
    """An abstract class that defines the interface to a original delete query API.

    .. note::

       This interface will eventually be deprecated when the new :class:`Query
       API<Query>` is finalised.

    ``Delete`` query objects cannot be constructed directly. Instead a
    ``Delete`` object is returned by the ``FactBase.delete()`` function. Given a
    ``FactBase`` object ``fb``, a specification is of the form:

          ``query = fb.delete(<predicate>).where(<expression>)``

    where ``<predicate>`` specifies the predicate type to search for,
    ``<expression>`` specifies the search criteria. The ``where()`` clause can
    be omitted in which case all predicates of that type will be deleted.

    """

    @abc.abstractmethod
    def where(self, *expressions):
        """Set the select statement's where clause.

        See the documentation for ``Select.where()`` for further details.
        """
        pass

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """Function to execute the delete query"""
        pass


# ------------------------------------------------------------------------------
# Query API version 1 with new query engine
# ------------------------------------------------------------------------------


class SelectImpl(Select):
    def __init__(self, factbase, qspec):
        self._factbase = factbase
        self._qspec = qspec

    # --------------------------------------------------------------------------
    # Add an order_by expression
    # --------------------------------------------------------------------------
    def where(self, *expressions):
        if self._qspec.where:
            raise TypeError("Cannot specify 'where' multiple times")
        if not expressions:
            raise TypeError("Empty 'where' expression")

        try:
            if len(expressions) == 1:
                where = process_where(expressions[0], self._qspec.roots)
            else:
                where = process_where(and_(*expressions), self._qspec.roots)
            nqspec = self._qspec.newp(where=where)
        except ValueError as e:
            raise TypeError(str(e)) from None
        return SelectImpl(self._factbase, nqspec)

    # --------------------------------------------------------------------------
    # Add an order_by expression
    # --------------------------------------------------------------------------
    def order_by(self, *expressions):
        if self._qspec.order_by:
            raise TypeError("Cannot specify 'order_by' multiple times")
        if not expressions:
            raise TypeError("Empty 'order_by' expression")
        try:
            order_by = process_orderby(expressions, self._qspec.roots)
            nqspec = self._qspec.newp(order_by=order_by)
        except ValueError as e:
            raise TypeError(str(e)) from None
        return SelectImpl(self._factbase, nqspec)

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def query_plan(self, *args, **kwargs):
        qspec = self._qspec.fill_defaults()

        (factsets, factindexes) = QueryExecutor.get_factmap_data(self._factbase.factmaps, qspec)
        qplan = make_query_plan(factindexes.keys(), qspec)

        return qplan.ground(*args, **kwargs)

    # --------------------------------------------------------------------------
    # Functions currently mirroring the old interface
    # --------------------------------------------------------------------------

    def get(self, *args, **kwargs):
        qspec = self._qspec
        if args or kwargs:
            if self._qspec.where is None:
                raise ValueError(("No where clause to ground"))
            qspec = self._qspec.bindp(*args, **kwargs)

        qe = QueryExecutor(self._factbase.factmaps, qspec)
        return list(qe.all())

    def get_unique(self, *args, **kwargs):
        qspec = self._qspec
        if args or kwargs:
            if self._qspec.where is None:
                raise ValueError(("No where clause to ground"))
            qspec = self._qspec.bindp(*args, **kwargs)

        qe = QueryExecutor(self._factbase.factmaps, qspec)
        found = None
        for out in qe.all():
            if found:
                raise ValueError("Query returned more than a single element")
            found = out
        return found

    def count(self, *args, **kwargs):
        qspec = self._qspec
        if args or kwargs:
            if self._qspec.where is None:
                raise ValueError(("No where clause to ground"))
            qspec = self._qspec.bindp(*args, **kwargs)

        qe = QueryExecutor(self._factbase.factmaps, qspec)
        count = 0
        for _ in qe.all():
            count += 1
        return count


# ------------------------------------------------------------------------------
# The Delete class
# ------------------------------------------------------------------------------


class _Delete(Delete):
    def __init__(self, factbase, qspec):
        self._factbase = factbase
        self._root = qspec.roots[0]
        self._select = SelectImpl(factbase, qspec)
        self._has_where = False

    def where(self, *expressions):
        self._has_where = True
        self._select = self._select.where(*expressions)
        return self

    def execute(self, *args, **kwargs):
        factmap = self._factbase.factmaps[self._root.meta.predicate]

        # If there is no where clause then delete everything
        if not self._has_where:
            num_deleted = len(factmap.facts())
            factmap.clear()
            return num_deleted

        # Gather all the facts to delete and remove them
        to_delete = [f for f in self._select.get(*args, **kwargs)]
        for fact in to_delete:
            factmap.remove(fact)
        return len(to_delete)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



"""A library of Python bool functions and terms for use within an ASP
program.

"""

from __future__ import annotations

from typing import Any

from clorm.orm.core import AnySymbol, BaseField, IntegerField
from clorm.orm.noclingo import Number

BOOL_FALSE = {0, "0", "off", "f", "false", "n", "no"}
BOOL_TRUE = {1, "1", "on", "t", "true", "y", "yes"}


def bool_validator(v: Any) -> bool:
    if v is True or v is False:
        return v
    if isinstance(v, str):
        v = v.lower()
    if v in BOOL_TRUE:
        return True
    if v in BOOL_FALSE:
        return False
    raise TypeError(f"value '{v}' could not be parsed to a boolean")


class BooleanField(BaseField):
    """A field to convert between a Clingo.Number/String object and a Python bool.

    Conversion failes if the value/clingo.Symbol is not one of the following:
    - a valid boolean ('True' or 'False')
    - the integers '0' or '1'
    - a 'str' which when converted to lower case in one of '0', 'off', 'f', 'false', 'n', 'no', '1', 'on', 't', 'true', 'y', 'yes'
    """

    def cltopy(symbol: AnySymbol) -> bool:
        try:
            return bool_validator(symbol.number)
        except (AttributeError, RuntimeError):
            pass
        try:
            return bool_validator(symbol.string)
        except (AttributeError, RuntimeError):
            pass
        raise TypeError(
            ("Object '{}' ({}) is not a Number/String " "Symbol").format(symbol, type(symbol))
        )

    def pytocl(v):
        val = int(bool_validator(v))
        return Number(val)


class StrictBooleanField(IntegerField):
    """A field to convert between a Clingo.Number object and a Python bool
    - Clingo.Number(0) will be converted to 'False'
    - Clingo.Number(1) will be converted to 'True'

    For everything else the conversion fails
    """

    def cltopy(symbol) -> bool:
        if symbol == 1:
            return True
        if symbol == 0:
            return False
        raise TypeError("value must be either '0' or '1'")

    def pytocl(v):
        if isinstance(v, bool):
            return int(v)
        raise TypeError("value is not a valid boolean")



"""contains clorm specific types"""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Tuple, TypeVar

__all__ = [
    "ConstantStr",
    "HeadList",
    "HeadListReversed",
    "StrictBool",
    "TailList",
    "TailListReversed",
]

if TYPE_CHECKING:
    ConstantStr = str
else:

    class ConstantStr(str):
        """A special ``str`` sub-class for specifying constant logical terms."""

        pass


_T = TypeVar("_T")

if TYPE_CHECKING:
    # perhaps renaming to something like Head(First)Tuple(Reversed), Tail(First)Tuple(Reversed)
    HeadList = Tuple[_T, ...]
    HeadListReversed = Tuple[_T, ...]
    TailList = Tuple[_T, ...]
    TailListReversed = Tuple[_T, ...]
else:

    class HeadList(Generic[_T]):
        pass

    class HeadListReversed(Generic[_T]):
        pass

    class TailList(Generic[_T]):
        pass

    class TailListReversed(Generic[_T]):
        pass


if TYPE_CHECKING:
    StrictBool = bool
else:

    class StrictBool(int):
        """
        StrictBool to allow for bools which are not type-coerced.
        """

        pass



"""A plugin replacement for the ``clingo`` modules that wraps the key clingo
objects to integrate Clorm ``Predicate`` and ``FactBase`` interfaces. See the
`Clingo API <https://potassco.org/clingo/python-api/current/clingo.html#Model>`_
for more details.

"""

from __future__ import annotations

import clingo as oclingo

# ------------------------------------------------------------------------------
# Reference to the original clingo objects so that when we replace them our
# references point to the originals.
# ------------------------------------------------------------------------------
from .orm import *

# I want to replace the original clingo - re-exporting everything in clingo
# except replacing the class overides with my version: _class_overides = [
# 'Control', 'Model', 'SolveHandle' ]. The following seems to work but I'm not
# sure if this is bad.


if oclingo.__version__ >= "5.5.0":
    from clingo.ast import parse_string
else:
    from clingo import parse_program  # type: ignore

from clingo import *

# export clorm's redefinitions of Control, Model and SolveHandle twice,
# once how they are defined and once with original name (clingo)
# this give the possibility to either replace the original classes with clorm's ones
# or use clorm's classes explicitly which also works better with static type checkers
from ._clingo import ClormControl as Control
from ._clingo import ClormModel as Model
from ._clingo import ClormSolveHandle as SolveHandle
from ._clingo import *

__all__ = list([k for k in oclingo.__dict__.keys() if k[0] != "_"])

__version__ = oclingo.__version__



# ------------------------------------------------------------------------------
# Unit tests for the peewee based data model
# ------------------------------------------------------------------------------

import datetime
import unittest

from clorm import IntegerField, clingo
from clorm.lib.date import (
    DateField,
    EnumDate,
    EnumDateRange,
    cl_date_range,
    cl_dow,
    date_range,
    dow,
)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class LibDateTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    # Make sure DateField does what we expect
    # --------------------------------------------------------------------------
    def test_datefield(self):
        dt = datetime.date(2017, 5, 16)
        raw = clingo.String("2017-05-16")
        self.assertEqual(DateField.pytocl(dt), raw)
        self.assertEqual(DateField.cltopy(raw), dt)

    # --------------------------------------------------------------------------
    # Make sure EnumDate does what we expect
    # --------------------------------------------------------------------------
    def test_enumdate(self):
        rawstr = clingo.String("2017-05-16")
        rawnum = clingo.Number(1)
        rawtuple = clingo.Function("", [rawnum, rawstr])
        ed = EnumDate(idx=1, date=datetime.date(2017, 5, 16))

        self.assertEqual(EnumDate.Field.pytocl(ed), rawtuple)
        self.assertEqual(EnumDate.Field.cltopy(rawtuple), ed)

    # --------------------------------------------------------------------------
    # Make sure the date functions behave correctly
    # --------------------------------------------------------------------------
    def test_dates(self):

        firstdate = datetime.date(2018, 1, 1)
        lastdate = datetime.date(2018, 1, 5)

        dates1 = [
            datetime.date(2018, 1, 1),
            datetime.date(2018, 1, 2),
            datetime.date(2018, 1, 3),
            datetime.date(2018, 1, 4),
        ]
        self.assertEqual(date_range(firstdate, lastdate), dates1)

        dates2 = [datetime.date(2018, 1, 1), datetime.date(2018, 1, 3)]
        self.assertEqual(date_range(firstdate, lastdate, 2), dates2)

        cl_firstdate = DateField.pytocl(firstdate)
        cl_lastdate = DateField.pytocl(lastdate)
        cl_2 = IntegerField.pytocl(2)

        cl_dates1 = [DateField.pytocl(d) for d in dates1]
        cl_dates2 = [DateField.pytocl(d) for d in dates2]

        self.assertEqual(cl_date_range(cl_firstdate, cl_lastdate), cl_dates1)
        self.assertEqual(cl_date_range(cl_firstdate, cl_lastdate, cl_2), cl_dates2)

        # Test the day-of-week function
        self.assertEqual(dow(firstdate), "monday")
        self.assertEqual(cl_dow(cl_firstdate), clingo.Function("monday", []))

    # --------------------------------------------------------------------------
    # Make sure the enumdate functions behave correctly
    # --------------------------------------------------------------------------
    def test_enumdates(self):

        firstdate = datetime.date(2018, 1, 1)
        lastdate = datetime.date(2018, 1, 5)
        datecount = 4

        dates1 = [
            EnumDate(idx=0, date=datetime.date(2018, 1, 1)),
            EnumDate(idx=1, date=datetime.date(2018, 1, 2)),
            EnumDate(idx=2, date=datetime.date(2018, 1, 3)),
            EnumDate(idx=3, date=datetime.date(2018, 1, 4)),
        ]
        dates2 = [
            EnumDate(idx=0, date=datetime.date(2018, 1, 1)),
            EnumDate(idx=1, date=datetime.date(2018, 1, 3)),
        ]

        cl_firstdate = DateField.pytocl(firstdate)
        cl_lastdate = DateField.pytocl(lastdate)
        cl_2 = IntegerField.pytocl(2)
        cl_4 = IntegerField.pytocl(4)

        edr1 = EnumDateRange(start=firstdate, stop=lastdate)
        edr2 = EnumDateRange(start=firstdate, count=2, step=2)

        self.assertEqual(edr1.enumdate_range(), dates1)
        self.assertEqual(edr2.enumdate_range(), dates2)
        self.assertEqual(edr1.last(), dates1[-1])

        rawdates1 = [EnumDate.Field.pytocl(ed) for ed in dates1]
        rawdates2 = [EnumDate.Field.pytocl(ed) for ed in dates2]

        self.assertEqual(edr1.cl_enumdate_range(), rawdates1)
        self.assertEqual(edr2.cl_enumdate_range(), rawdates2)

        # Test the day-of-week function
        self.assertEqual(edr1.dow(dates1[0]), "monday")
        self.assertEqual(edr1.cl_dow(rawdates1[0]), clingo.Function("monday", []))

    # --------------------------------------------------------------------------
    # Check that the docstring are the same
    # --------------------------------------------------------------------------
    def test_docstrings(self):
        self.assertEqual(dow.__doc__, cl_dow.__doc__)
        self.assertEqual(date_range.__doc__, cl_date_range.__doc__)

        Edr = EnumDateRange
        self.assertEqual(Edr.dow.__doc__, Edr.cl_dow.__doc__)
        self.assertEqual(Edr.enumdate_range.__doc__, Edr.cl_enumdate_range.__doc__)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



# ------------------------------------------------------------------------------
# Unit tests for the clorm monkey patching
# ------------------------------------------------------------------------------
import unittest

from clorm.util.wrapper import WrapperMetaClass, init_wrapper, make_class_wrapper

from .support import check_errmsg

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class WrapperMetaClassTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_simple_wrapper(self):
        class My(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            # A getter only
            @property
            def num1(self):
                return self._num

            # A getter and setter
            @property
            def num2(self):
                return self._num

            @num2.setter
            def num2(self, x):
                self._num = x

        class MyWrapper(My, metaclass=WrapperMetaClass):
            pass

        # Despite the specification MyWrapper is a subclass of object and not My
        self.assertEqual(len(MyWrapper.__bases__), 1)
        self.assertEqual(MyWrapper.__bases__[0], object)

        # Use the wrapper to create a new object
        mypr = MyWrapper(5)
        self.assertEqual(mypr.num1, 5)

        # Use the wrapper on an existing object
        my = My(5)
        mypr = MyWrapper(wrapped_=my)
        self.assertEqual(mypr.num1, 5)
        mypr.inc()
        self.assertEqual(mypr.num1, 6)
        self.assertEqual(mypr.__doc__, my.__doc__)

        # Test that the num2 setter is also wrapped
        mypr.num2 = 2
        self.assertEqual(my.num2, 2)
        self.assertEqual(my.num2, mypr.num1)

        # Test that trying to set the getter only num1 fails
        with self.assertRaises(AttributeError) as ctx:
            mypr.num1 = 5

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_wrapper_with_overrides(self):
        class My(object):
            """My Class"""

            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

            def __str__(self):
                return "My({})".format(self._num)

        class MyWrapper(My, metaclass=WrapperMetaClass):
            def __init__(self, *args, **kwargs):
                if "my_" in kwargs:
                    if len(args) != 0 and len(kwargs) != 1:
                        raise ValueError("Missing 'my_' argument ")
                    init_wrapper(self, wrapped_=kwargs["my_"])
                else:
                    init_wrapper(self, *args, **kwargs)

            @property
            def my_(self):
                return self._wrapped

            def inc(self):
                self._wrapped.inc()
                self._wrapped.inc()

        # Use the wrapper on an existing object
        my = My(5)
        mypr = MyWrapper(my_=my)
        self.assertEqual(mypr.my_, my)
        self.assertEqual(mypr.num, 5)
        mypr.inc()
        self.assertEqual(mypr.num, 7)
        self.assertEqual(str(mypr), str(my))
        self.assertEqual(mypr.__doc__, my.__doc__)
        self.assertNotEqual(repr(mypr), repr(my))

        # Use the wrapper to create a new object
        mypr = MyWrapper(5)
        self.assertEqual(type(mypr.my_), My)
        self.assertEqual(mypr.num, 5)

    # --------------------------------------------------------------------------
    # Test that we can actually use an object of a different type instead of the
    # officially wrapped type. Allows a form of duck-typing.
    # --------------------------------------------------------------------------
    def test_wrapper_diff_type(self):
        class My1(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

        class My2(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 2

            @property
            def num(self):
                return self._num + 1

        class MyWrapper(My1, metaclass=WrapperMetaClass):
            pass

        ok1 = My1(5)
        ok2 = My2(5)
        wrapper1 = MyWrapper(wrapped_=ok1)
        wrapper2 = MyWrapper(wrapped_=ok2)

        self.assertEqual(wrapper1.num, 5)
        self.assertEqual(wrapper2.num, 6)
        wrapper1.inc()
        wrapper2.inc()
        self.assertEqual(wrapper1.num, 6)
        self.assertEqual(wrapper2.num, 8)

    # --------------------------------------------------------------------------
    # Test that we can wrap a wrapper.
    # --------------------------------------------------------------------------
    def test_wrapper_of_wrapper(self):
        class My(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

        class MyWrapper(My, metaclass=WrapperMetaClass):
            def inc(self):
                self._wrapped.inc()
                self._wrapped.inc()

            @property
            def num(self):
                return self._wrapped.num + 1

        my1 = My(5)
        my2 = MyWrapper(wrapped_=my1)
        my3 = MyWrapper(wrapped_=my2)

        self.assertEqual(my1.num, 5)
        self.assertEqual(my2.num, 6)
        self.assertEqual(my3.num, 7)

        # Incrementing the underlying object
        my1.inc()
        self.assertEqual(my1.num, 6)
        self.assertEqual(my2.num, 7)
        self.assertEqual(my3.num, 8)

        # Incrementing the first level wrapper
        my2.inc()
        self.assertEqual(my1.num, 8)
        self.assertEqual(my2.num, 9)
        self.assertEqual(my3.num, 10)

        # Incrementing the second level wrapper
        my3.inc()
        self.assertEqual(my1.num, 12)
        self.assertEqual(my2.num, 13)
        self.assertEqual(my3.num, 14)

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_bad_wrapper(self):
        class MyClass(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

        class Other(object):
            def __init__(self, num):
                self._num2 = num

        # Multiple inheritence is not allowed
        with self.assertRaises(TypeError) as ctx:

            class MyWrapper(MyClass, Other, metaclass=WrapperMetaClass):
                pass

        check_errmsg("ProxyMetaClass requires exactly one", ctx)

        # Using an object of the wrong class with missing attribute
        with self.assertRaises(AttributeError) as ctx:

            class MyWrapper(MyClass, metaclass=WrapperMetaClass):
                pass

            other = Other(4)
            my = MyWrapper(wrapped_=other)
            my.num
        check_errmsg("'Other' object has no attribute 'num'", ctx)


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


class WrapperFunctionTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_simple_wrapper(self):
        class My(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            # A getter only
            @property
            def num1(self):
                return self._num

            # A getter and setter
            @property
            def num2(self):
                return self._num

            @num2.setter
            def num2(self, x):
                self._num = x

        class MyWrapperT(object):
            pass

        MyWrapper = make_class_wrapper(My, MyWrapperT)

        # Despite the specification MyWrapper is a subclass of object and not My
        self.assertEqual(len(MyWrapper.__bases__), 1)
        self.assertEqual(MyWrapper.__bases__[0], object)

        # Use the wrapper to create a new object
        mypr = MyWrapper(5)
        self.assertEqual(mypr.num1, 5)

        # Use the wrapper on an existing object
        my = My(5)
        mypr = MyWrapper(wrapped_=my)
        self.assertEqual(mypr.num1, 5)
        mypr.inc()
        self.assertEqual(mypr.num1, 6)
        self.assertEqual(mypr.__doc__, my.__doc__)

        # Test that the num2 setter is also wrapped
        mypr.num2 = 2
        self.assertEqual(my.num2, 2)
        self.assertEqual(my.num2, mypr.num1)

        # Test that trying to set the getter only num1 fails
        with self.assertRaises(AttributeError) as ctx:
            mypr.num1 = 5

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_wrapper_with_overrides(self):
        class My(object):
            """My Class"""

            def __init__(self, num):
                self._num = num

            def inc(self):
                """Inc"""
                self._num += 1

            @property
            def num(self):
                """Num"""
                return self._num

            def __str__(self):
                return "My({})".format(self._num)

        class MyOverride(object):
            def __init__(self, *args, **kwargs):
                if "my_" in kwargs:
                    if len(args) != 0 and len(kwargs) != 1:
                        raise ValueError("Missing 'my_' argument ")
                    init_wrapper(self, wrapped_=kwargs["my_"])
                else:
                    init_wrapper(self, *args, **kwargs)

            @property
            def my_(self):
                return self._wrapped

            def inc(self):
                self._wrapped.inc()
                self._wrapped.inc()

        MyWrapper = make_class_wrapper(My, MyOverride)

        self.assertEqual(MyWrapper.__doc__, My.__doc__)
        self.assertEqual(MyWrapper.num.__doc__, My.num.__doc__)
        self.assertEqual(MyWrapper.inc.__doc__, My.inc.__doc__)
        # Use the wrapper on an existing object
        my = My(5)
        mypr = MyWrapper(my_=my)
        self.assertEqual(mypr.my_, my)
        self.assertEqual(mypr.num, 5)
        mypr.inc()
        self.assertEqual(mypr.num, 7)
        self.assertEqual(str(mypr), str(my))
        self.assertEqual(mypr.__doc__, my.__doc__)
        self.assertNotEqual(repr(mypr), repr(my))

        # Use the wrapper to create a new object
        mypr = MyWrapper(5)
        self.assertEqual(type(mypr.my_), My)
        self.assertEqual(mypr.num, 5)

    # --------------------------------------------------------------------------
    # Test that we can actually use an object of a different type instead of the
    # officially wrapped type. Allows a form of duck-typing.
    # --------------------------------------------------------------------------
    def test_wrapper_diff_type(self):
        class My1(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

        class My2(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 2

            @property
            def num(self):
                return self._num + 1

        MyWrapper = make_class_wrapper(My1)

        ok1 = My1(5)
        ok2 = My2(5)
        wrapper1 = MyWrapper(wrapped_=ok1)
        wrapper2 = MyWrapper(wrapped_=ok2)

        self.assertEqual(wrapper1.num, 5)
        self.assertEqual(wrapper2.num, 6)
        wrapper1.inc()
        wrapper2.inc()
        self.assertEqual(wrapper1.num, 6)
        self.assertEqual(wrapper2.num, 8)

    # --------------------------------------------------------------------------
    # Test that we can wrap a wrapper.
    # --------------------------------------------------------------------------
    def test_wrapper_of_wrapper(self):
        class My(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

        class MyOverride(object):
            def inc(self):
                self._wrapped.inc()
                self._wrapped.inc()

            @property
            def num(self):
                return self._wrapped.num + 1

        MyWrapper = make_class_wrapper(My, MyOverride)

        my1 = My(5)
        my2 = MyWrapper(wrapped_=my1)
        my3 = MyWrapper(wrapped_=my2)

        self.assertEqual(my1.num, 5)
        self.assertEqual(my2.num, 6)
        self.assertEqual(my3.num, 7)

        # Incrementing the underlying object
        my1.inc()
        self.assertEqual(my1.num, 6)
        self.assertEqual(my2.num, 7)
        self.assertEqual(my3.num, 8)

        # Incrementing the first level wrapper
        my2.inc()
        self.assertEqual(my1.num, 8)
        self.assertEqual(my2.num, 9)
        self.assertEqual(my3.num, 10)

        # Incrementing the second level wrapper
        my3.inc()
        self.assertEqual(my1.num, 12)
        self.assertEqual(my2.num, 13)
        self.assertEqual(my3.num, 14)

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_bad_wrapper(self):
        class MyClass(object):
            def __init__(self, num):
                self._num = num

            def inc(self):
                self._num += 1

            @property
            def num(self):
                return self._num

        class Other(object):
            def __init__(self, num):
                self._num2 = num

        # Using an object of the wrong class with missing attribute
        with self.assertRaises(AttributeError) as ctx:
            MyWrapper = make_class_wrapper(MyClass)
            other = Other(4)
            my = MyWrapper(wrapped_=other)
            my.num
        check_errmsg("'Other' object has no attribute 'num'", ctx)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



from __future__ import annotations

import inspect
import sys
from inspect import FrameInfo
from typing import (
    Any,
    Dict,
    ForwardRef,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    _eval_type,
    cast,
    get_args,
    get_origin,
)

from clingo import Symbol

from .noclingo import NoSymbol

_T0 = TypeVar("_T0", bound=Any)
_T1 = TypeVar("_T1", bound=Any)
_T2 = TypeVar("_T2", bound=Any)
_T3 = TypeVar("_T3", bound=Any)
_T4 = TypeVar("_T4", bound=Any)


AnySymbol = Union[Symbol, NoSymbol]


def print_dict(adict):
    for k, v in adict.items():
        print(f"{k} => {v}")


def resolve_annotations(
    raw_annotations: Dict[str, Type[Any]],
    module_name: Optional[str] = None,
    locals_: Dict[str, Any] = None,
) -> Dict[str, Type[Any]]:
    """
    Taken from https://github.com/pydantic/pydantic/blob/1.10.X-fixes/pydantic/typing.py#L376
    with some modifications for handling when the first _eval_type() call fails.

    Resolve string or ForwardRef annotations into type objects if possible.
    """
    base_locals: Dict[str, Type[Any]] = dict(locals_) if locals_ else {}
    base_globals: Optional[Dict[str, Any]] = None
    if module_name:
        try:
            module = sys.modules[module_name]
        except KeyError:
            # happens occasionally, see https://github.com/pydantic/pydantic/issues/2363
            pass
        else:
            base_globals = module.__dict__

    annotations = {}
    frameinfos: Union[list[FrameInfo], None] = None
    for name, value in raw_annotations.items():
        if isinstance(value, str):

            # Turn the string type annotation into a ForwardRef for processing
            if (3, 10) > sys.version_info >= (3, 9, 8) or sys.version_info >= (3, 10, 1):
                value = ForwardRef(value, is_argument=False, is_class=True)
            else:
                value = ForwardRef(value, is_argument=False)
        try:
            if sys.version_info < (3, 13):
                type_ = _eval_type(value, base_globals, base_locals)
            else:
                type_ = _eval_type(value, base_globals, base_locals, type_params=())

        except NameError:
            # The type annotation could refer to a definition at a non-global scope so build
            # the locals from the calling context. We reuse the same set of locals for
            # multiple annotations.
            if frameinfos is None:
                frameinfos = inspect.stack()
                if len(frameinfos) < 4:
                    raise RuntimeError(
                        'Cannot resolve field "{name}" with type annotation "{value}"'
                    )
                frameinfos = frameinfos[3:]
            type_ = None
            count = 0
            while frameinfos:
                try:
                    type_ = _eval_type(value, base_globals, base_locals)
                    break
                except NameError:
                    finfo = frameinfos.pop(0)
                    base_locals.update(finfo.frame.f_locals)
            if type_ is None:
                raise RuntimeError(
                    f'Cannot resolve field "{name}" with type annotation "{value}"'
                )
        annotations[name] = type_
    return annotations
